/*
 * band_cholesky.c --- оӹ A  Cholesky ʬ
 */

#include <stdio.h>
#include <limits.h>
#include <math.h>
#include <matrix.h>

double drandom() { return random() / (double) INT_MAX; }
double sqr(double x) { return x * x; }
double max(double a, double b) { return (a > b) ? a : b; }
double min(double a, double b) { return (a < b) ? a : b; }

/*  Cholesky ʬ A=U^T U (A, U гξȾʬΤͿ)
 *
 * ȾХ m  n оӹ A=(a_{ij}) (1in, 1jn) 
 *   ASB=(asb_{ij) (1in, 1jm+1)
 * ˵ͤᤳ
 *
 *   asb[i][j-i+1] = a[i][j]  (1in, ijmin(i+m,n))
 *
 * ̤ΰ U  ASB ˾񤭤롣
 *
 */
void band_cholesky(int n, int m, matrix ASB)
{
  int j, i, k;
  double s;
  for (i = 1; i <= n; i++) {
    /* U[i][i] η׻ */
    s = ASB[i][1];
    for (k = max(1,i-m); k < i; k++)
      s -= sqr(ASB[k][i-k+1]);
    if (s < 0) {
      fprintf(stderr, "s < 0\n");
      exit(0);
    }
    ASB[i][1] = sqrt(s);
    /* U[i][j] (i < j) η׻ */
    for (j = i + 1; j <= min(i + m, n); j++) {
      s = ASB[i][j-i+1];
      for (k = max(1,j-m); k < i; k++)
        s -= ASB[k][j-k+1] * ASB[k][i-k+1];
      ASB[i][j-i+1] = s / ASB[i][1];
    }
  }
}

/* ѹ򷸿ȤϢΩ켡 */
/* U^T x=b  */
void solve_band_utxb(int n, int m, matrix UB, vector b)
{
  int i, j;
  for (i = 1; i <= n; i++) {
    for (j = max(i-m,1); j < i; j++)
      b[i] -= UB[j][i-j+1] * b[j];
    b[i] /= UB[i][1];
  }
}

/* Ux=b  */
void solve_band_uxb(int n, int m, matrix UB, vector b)
{
  int i, j;
  for (i = n; i >= 1; i--) {
    for (j = i + 1; j <= min(i+m,n); j++)
      b[i] -= UB[i][j-i+1] * b[j];
    b[i] /= UB[i][1];
  }
}

/* ѥåƤӹɽ */
void print_symmetric_band_matrix1(int n, int m, matrix asb)
{
  int i, j;
  for (i = 1; i <= n; i++) {
    for (j = 1; j <= n; j++)
       if (abs(j - i) > m)
         printf("%f ", 0.0);
       else {
         if (j >= i)
           printf("%f ", asb[i][j-i+1]);
         else /* פ i  j ؤ */
           printf("%f ", asb[j][i-j+1]);
       }
    printf("\n");
  }
}

/* ͤ廰ӹɽ */
void print_upper_triangular_band_matrix1(int n, int m, matrix ub)
{
  int i, j;
  for (i = 1; i <= n; i++) {
    for (j = 1; j <= n; j++)
       if (abs(j - i) > m)
         printf("%f ", 0.0);
       else {
         if (j >= i)
           printf("%f ", ub[i][j-i+1]);
         else
           printf("%f ", 0.0);
       }
    printf("\n");
  }
}

void print_vector1(int n, vector x)
{
  int i;
  for (i = 1; i <= n; i++)
    printf("%f ", x[i]);
  printf("\n");
}

int main()
{
  int n, m, i, j, k;
  double s;
  matrix ASB, UB;
  vector x, b;

  /* ̤οθĿ, ȾХ */
  n = 8; m = 3;

  /* ASB, UB  n(m+1) ι˵ͤ */
  ASB = new_matrix(n + 1, (m+1) + 1);
  UB = new_matrix(n + 1, (m+1) + 1);
  x = new_vector(n + 1);
  b = new_vector(n + 1);

  /* 廰ѹ U */
  for (i = 1; i <= n; i++)
    for (j = i; j <= min(i+m, n); j++)
      UB[i][j-i+1] = drandom();
  printf("U=\n");
  print_upper_triangular_band_matrix1(n, m, UB);

  /* A=U^T U ׻롣оΤʤΤгξ¦׻롣 */
  for (i = 1; i <= n; i++)
    for (j = i; j <= min(i+m,n); j++) { /* гξ */
      s = 0;
      /* U[k][i]*U[k][j]
         max(i-m,1)ki  max(j-m,1)kj ξΤ 0
         Ȥ i  j  max(j-m,1)ki Ȥ
       */
      for (k = max(j-m,1); k <= i; k++)
        s += UB[k][i-k+1] * UB[k][j-k+1];
      ASB[i][j-i+1] = s;
    }
  printf("A:=U^T U\n");
  print_symmetric_band_matrix1(n, m, ASB);

  /* ñ x */
  for (i = 1; i <= n; i++)
    x[i] = i;

  /* b=A x */
  for (i = 1; i <= n; i++) {
    s = ASB[i][1] * x[i];
    for (j = i + 1; j <= min(i+m,n); j++)
      s += ASB[i][j-i+1] * x[j];
    for (j = max(i-m,1); j < i; j++)
      s += ASB[j][i-j+1] * x[j];
    b[i] = s;
  }

  /* Cholesky ʬ A=U^T U */
  band_cholesky(n, m, ASB);
  printf("Cholesky ʬη U=\n");
  print_upper_triangular_band_matrix1(n, m, ASB);

  /* U^T U x = b  */
  solve_band_utxb(n, m, UB, b);
  solve_band_uxb(n, m, UB, b);
  printf("x=\n");
  print_vector1(n, b);

  return 0;
}
