/*
 * band_cholesky_C++.C --- оӹ A  Cholesky ʬ
 */

#include <iostream.h>
#include <iomanip.h>
#include <limits.h>
#include <math.h>
#include <Matrix.h>
#include <Vector.h>

double drandom(void) { return random() / (double) INT_MAX; }
double sqr(double x) { return x * x; }
// double max(double a, double b) { return (a > b) ? a : b; }
// double min(double a, double b) { return (a < b) ? a : b; }

/*  Cholesky ʬ A=U^T U (A, U гξȾʬΤͿ)
 *
 * ȾХ m  n оӹ A=(a_{ij}) (1in, 1jn) 
 *   ASB=(asb_{ij) (1in, 1jm+1)
 * ˵ͤᤳ
 *
 *   asb[i][j-i+1] = a[i][j]  (1in, ijmin(i+m,n))
 *
 * ̤ΰ U  ASB ˾񤭤롣
 *
 */
void band_cholesky(int n, int m, MATRIX &ASB)
{
  double s;
  for (int i = 1; i <= n; i++) {
    /* U[i][i] η׻ */
    s = ASB(i,1);
    for (int k = max(1,i-m); k < i; k++)
      s -= sqr(ASB(k,i-k+1));
    if (s < 0) {
      cerr << "s < 0" << endl;
      exit(0);
    }
    ASB(i,1) = sqrt(s);
    /* U[i][j] (i < j) η׻ */
    for (int j = i + 1; j <= min(i + m, n); j++) {
      s = ASB(i,j-i+1);
      for (int k = max(1,j-m); k < i; k++)
        s -= ASB(k,j-k+1) * ASB(k,i-k+1);
      ASB(i,j-i+1) = s / ASB(i,1);
    }
  }
}

/* ѹ򷸿ȤϢΩ켡 */
/* U^T x=b  */
void solve_band_utxb(int n, int m, const MATRIX &UB, VECTOR &b)
{
  for (int i = 1; i <= n; i++) {
    for (int j = max(i-m,1); j < i; j++)
      b(i) -= UB(j,i-j+1) * b(j);
    b(i) /= UB(i,1);
  }
}

/* Ux=b  */
void solve_band_uxb(int n, int m, const MATRIX &UB, VECTOR &b)
{
  for (int i = n; i >= 1; i--) {
    for (int j = i + 1; j <= min(i+m,n); j++)
      b(i) -= UB(i,j-i+1) * b(j);
    b(i) /= UB(i,1);
  }
}

/* ѥåƤӹɽ */
void print_symmetric_band_matrix1(int n, int m, const MATRIX &asb)
{
  for (int i = 1; i <= n; i++) {
    for (int j = 1; j <= n; j++)
       if (abs(j - i) > m)
         cout << 0.0 << " ";
       else {
         if (j >= i)
	   cout << asb(i,j-i+1) << " ";
         else /* פ i  j ؤ */
           cout << asb(j,i-j+1) << " ";
       }
    cout << endl;
  }
}

/* ͤ廰ӹɽ */
void print_upper_triangular_band_matrix1(int n, int m, const MATRIX &ub)
{
  for (int i = 1; i <= n; i++) {
    for (int j = 1; j <= n; j++)
       if (abs(j - i) > m)
         cout << 0.0 << " ";
       else {
         if (j >= i)
           cout << ub(i,j-i+1) << " ";
         else
           cout << 0.0 << " ";
       }
    cout << endl;
  }
}

void print_vector1(int n, const VECTOR &x)
{
  for (int i = 1; i <= n; i++)
    cout << x(i) << " ";
  cout << endl;
}

int main()
{
  int n, m, i, j, k;
  double s;

  cout << setprecision(4);
  cout << setiosflags(ios::fixed);
  /* ̤οθĿ, ȾХ */
  n = 8; m = 3;

  /* ASB, UB  n(m+1) ι˵ͤ */
  MATRIX ASB(n,m+1), UB(n,m+1); 
  VECTOR x(n), b(n);

  /* 廰ѹ U */
  for (i = 1; i <= n; i++)
    for (j = i; j <= min(i+m, n); j++)
      UB(i,j-i+1) = drandom();
  cout << "U=" << endl;
  print_upper_triangular_band_matrix1(n, m, UB);

  /* A=U^T U ׻롣оΤʤΤгξ¦׻롣 */
  for (i = 1; i <= n; i++)
    for (j = i; j <= min(i+m,n); j++) { /* гξ */
      s = 0;
      /* U[k](i)*U[k](j)
         max(i-m,1)ki  max(j-m,1)kj ξΤ 0
         Ȥ i  j  max(j-m,1)ki Ȥ
       */
      for (k = max(j-m,1); k <= i; k++)
        s += UB(k,i-k+1) * UB(k,j-k+1);
      ASB(i,j-i+1) = s;
    }
  cout << "A:=U^T U" << endl;
  print_symmetric_band_matrix1(n, m, ASB);

  /* ñ x */
  for (i = 1; i <= n; i++)
    x(i) = i;

  /* b=A x */
  for (i = 1; i <= n; i++) {
    s = ASB(i,1) * x(i);
    for (j = i + 1; j <= min(i+m,n); j++)
      s += ASB(i,j-i+1) * x(j);
    for (j = max(i-m,1); j < i; j++)
      s += ASB(j,i-j+1) * x(j);
    b(i) = s;
  }

  /* Cholesky ʬ A=U^T U */
  band_cholesky(n, m, ASB);
  cout << "Cholesky ʬη U=" << endl;
  print_upper_triangular_band_matrix1(n, m, ASB);

  /* U^T U x = b  */
  solve_band_utxb(n, m, UB, b);
  solve_band_uxb(n, m, UB, b);
  cout << "x=" << endl;
  print_vector1(n, b);

  return 0;
}
