/*
 * cholesky4.c
 */

#include <stdio.h>
#include <math.h>
#include <matrix.h>
#include <limits.h>

/* [0,1) ϰϤΰ͵ */
double drandom()
{
  return random() / (double) INT_MAX;
}

/* ߥ٥ȥ */
void mul_mv(int n, vector Ab, matrix A, vector b)
{
  int i;
  for (i = 1; i <= n; i++)
    Ab[i] = dotprod(n, A[i]+1, b+1);
}

/*  */
void mul_mm(int n, matrix ab, matrix a, matrix b)
{
  int i, j, k;
  double s;
  for (i = 1; i <= n; i++)
    for (j = 1; j <= n; j++) {
      s = 0;
      for (k = 1; k <= n; k++)
        s += a[i][k] * b[k][j];
      ab[i][j] = s;
    }
}

/* ٥ȥɽ */
void print_vector(int n, vector x)
{
  int i;
  for (i = 1; i <= n; i++)
    printf("%f ", x[i]);
  printf("\n");
}

/* ɽ */
void print_matrix(int n, matrix a)
{
  int i, j;
  for (i = 1; i <= n; i++) {
    for (j = 1; j <= n; j++)
      printf("%f ", a[i][j]);
    printf("\n");
  }
}

/* ˤ */
void clear_m(int n, matrix a)
{
  int i, j;
  for (i = 1; i <= n; i++)
    for (j = 1; j <= n; j++)
      a[i][j] = 0;
}

/*  */
double sqr(double x) { return x * x; }

/* Cholesky ʬ A=U^T U (A гξȾʬΤ߻q) */
void cholesky(int n, matrix U, matrix a)
{
  int j, i, k;
  double s;
  for (i = 1; i <= n; i++) {
    /* U[i][i] η׻ */
    s = a[i][i];
    for (k = 1; k < i; k++)
      s -= sqr(U[k][i]);
    if (s < 0) {
      fprintf(stderr, "s < 0\n");
      exit(0);
    }
    U[i][i] = sqrt(s);
    for (j = i + 1; j <= n; j++) {
      s = a[i][j];
      for (k = 1; k < i; k++)
        s -= U[k][j] * U[k][i];
      U[i][j] = s / U[i][i];
    }
  }
}

/* ѹ򷸿ȤϢΩ켡 */
/* Ux=b  */
void solve_uxb(int n, matrix u, vector b)
{
  int i, j;
  for (i = n; i >= 1; i--) {
    for (j = i + 1; j <= n; j++)
      b[i] -= u[i][j] * b[j];
    b[i] /= u[i][i];
  }
}

/* L^T x=b   solve_uxb()  u[i][j]  L[j][i]  */
void solve_ltxb(int n, matrix L, vector b)
{
  int i, j;
  for (i = n; i >= 1; i--) {
    for (j = i + 1; j <= n; j++)
      b[i] -= L[j][i] * b[j];
    b[i] /= L[i][i];
  }
}

/* Lx=b  */
void solve_lxb(int n, matrix L, vector b)
{
  int i, j;
  for (i = 1; i <= n; i++) {
    for (j = 1; j < i; j++)
      b[i] -= L[i][j] * b[j];
    b[i] /= L[i][i];
  }
}

/* U^T x=b  */
void solve_utxb(int n, matrix U, vector b)
{
  int i, j;
  for (i = 1; i <= n; i++) {
    for (j = 1; j < i; j++)
      b[i] -= U[j][i] * b[j];
    b[i] /= U[i][i];
  }
}

int main()
{
  int i, j, N;
  matrix a, U, Ut;
  vector x, b;
  printf("N="); scanf("%d", &N);
  a = new_matrix(N+1, N+1);
  U = new_matrix(N+1, N+1);
  Ut = new_matrix(N+1, N+1);
  x = new_vector(N+1);
  b = new_vector(N+1);
  for (i = 1; i <= N; i++) {
    for (j = 1; j < i; j++)
      Ut[j][i] = U[i][j] = 0.0;
    for (j = i; j <= N; j++)
      Ut[j][i] = U[i][j] = drandom();
  }
  mul_mm(N, a, Ut, U);
  printf("L=\n");
  print_matrix(N, U);
  printf("a=\n");
  print_matrix(N, a);

  clear_m(N, U);
  cholesky(N, U, a);
  printf("U=\n");
  print_matrix(N, U);

  /* ϢΩ1 */
  for (i = 1; i <= N; i++)
    x[i] = i;
  mul_mv(N, b, a, x);

  solve_utxb(N, U, b);
  solve_uxb(N, U, b);
  printf("x=\n");
  print_vector(N, b);

  return 0;
}
