/*
 * cholesky5.c
 *  ӹѤ Cholesky ʬ (A=U^T U)
 *  õ (빽Ѥʤ衢ν񤯤)
 */

#include <stdio.h>
#include <math.h>
#include <matrix.h>
#include <limits.h>

/* [0,1) ΰ͵ */
double drandom()
{
  return random() / (double) INT_MAX;
}

/* Ǿ,  */
int min(int a, int b) { return (a < b) ? a : b; }
int max(int a, int b) { return (a > b) ? a : b; }

/* Υꥢ */
void clear_matrix0(int m, int n, matrix a)
{
  int i, j;
  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
      a[i][j] = 0;
}

/* Υꥢ */
void clear_matrix1(int m, int n, matrix a)
{
  int i, j;
  for (i = 1; i <= m; i++)
    for (j = 1; j <= n; j++)
      a[i][j] = 0;
}

/* ӹ */
void random_band0(int n, int m, matrix a)
{
  int i, j;
  clear_matrix0(n, n, a);
  for (i = 0; i < n; i++)
    for (j = max(0, i-m); j <= min(i+m, n-1); j++)
      a[i][j] = drandom();
}

/* ӹ */
void random_band1(int n, int m, matrix a)
{
  int i, j;
  clear_matrix1(n, n, a);
  for (i = 1; i <= n; i++)
    for (j = max(1, i-m); j <= min(i+m, n); j++)
      a[i][j] = drandom();
}

/* оӹ */
void random_symmetric_band0(int n, int m, matrix a)
{
  int i, j;
  clear_matrix0(n, n, a);
  for (i = 0; i < n; i++)
    for (j = i; j <= min(i+m, n-1); j++)
      a[i][j] = a[j][i] = drandom();
}

/* оӹ */
void random_symmetric_band1(int n, int m, matrix a)
{
  int i, j;
  clear_matrix1(n, n, a);
  for (i = 1; i <= n; i++)
    for (j = i; j <= min(i+m, n); j++)
      a[i][j] = a[j][i] = drandom();
}

/* ɽ */
void print_matrix0(int m, int n, matrix a)
{
  int i, j;
  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++)
      printf("%f ", a[i][j]);
    printf("\n");
  }
}

/* ɽ */
void print_matrix1(int m, int n, matrix a)
{
  int i, j;
  for (i = 1; i <= m; i++) {
    for (j = 1; j <= n; j++)
      printf("%f ", a[i][j]);
    printf("\n");
  }
}

/* ٥ȥɽ */
void print_vector0(int n, vector x)
{
  int i;
  for (i = 0; i < n; i++)
    printf("%f ", x[i]);
  printf("\n");
}

/* ٥ȥɽ */
void print_vector1(int n, vector x)
{
  int i;
  for (i = 1; i <= n; i++)
    printf("%f ", x[i]);
  printf("\n");
}

/*
 * ӹͤ
 *
 * ȾХ m 
 *   A=(a_{ij}) (0in-1, 0jn-1)
 * 
 *   AB=(ab_{ij) (0in-1, 0j2m)
 * ˵ͤ
 *
 *   ab[i][j-i+m] = a[i][j]  (0in-1, max(i-m,0)jmin(i+m,n-1))
 *                                       |i-j|m Ȥ
 *
 *   a[i][i+j-m] = ab[i][j]  (0in-1, ??max(m-i,0)jmin(m-i+N-1,N-1))
 */
void pack0(int n, int m, matrix ab, matrix a)
{
  int i, j;
  clear_matrix0(n, 2*m+1, ab);
  for (i = 0; i < n; i++)
    for (j = max(i-m,0); j <= min(i+m,n-1); j++)
      ab[i][j-i+m] = a[i][j];
}

/*
 * ӹͤ
 *
 * ȾХ m 
 *   A=(a_{ij}) (1in, 1jn)
 * 
 *   AB=(ab_{ij) (1in, 1j2m+1)
 * ˵ͤ
 *
 *   ab[i][j-i+m+1] = a[i][j]  (1in, max(i-m,1)jmin(i+m,n))
 *                                       |i-j|m Ȥ
 *
 * a[i][i+j-m] = ab[i][j]  (0in-1, ??max(m-i,0)jmin(m-i+N-1,N-1))
 */
void pack1(int n, int m, matrix ab, matrix a)
{
  int i, j;
  clear_matrix1(n, 2*m+1, ab);
  for (i = 1; i <= n; i++)
    for (j = max(i-m,1); j <= min(i+m,n); j++)
      ab[i][j-i+m+1] = a[i][j];
}

/*
 * оӹͤ
 *
 * ȾХ m оι
 *   A=(a_{ij}) (0in-1, 0jn-1)
 * 
 *   ASB=(asb_{ij) (0in-1, 0jm)
 * ˵ͤ
 *
 *   asb[i][j-i] = a[i][j]  (0in-1, ijmin(i+m,n-1))
 *
 */
void pack_symmetric0(int n, int m, matrix asb, matrix a)
{
  int i, j;
  clear_matrix0(n, m+1, asb);
  for (i = 0; i < n; i++)
    for (j = i; j <= min(i+m,n-1); j++)
      asb[i][j-i] = a[i][j];
}

/*
 * оӹͤ
 *
 * ȾХ m оι
 *   A=(a_{ij}) (1in, 1jn)
 * 
 *   ASB=(asb_{ij) (1in, 1jm+1)
 * ˵ͤ
 *
 *   asb[i][j-i+1] = a[i][j]  (1in, ijmin(i+m,n))
 *
 */
void pack_symmetric1(int n, int m, matrix asb, matrix a)
{
  int i, j;
  clear_matrix1(n, m+1, asb);
  for (i = 1; i <= n; i++)
    for (j = i; j <= min(i+m,n); j++)
      asb[i][j-i+1] = a[i][j];
}

/* ѥåƤӹɽ */
void print_band_matrix0(int n, int m, matrix ab)
{
  int i, j;
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++)
       if (j < i - m || j > i + m)
         printf("%f ", 0.0);
       else
         printf("%f ", ab[i][j-i+m]);
    printf("\n");
  }
}

/* ѥåƤӹɽ */
void print_band_matrix1(int n, int m, matrix ab)
{
  int i, j;
  for (i = 1; i <= n; i++) {
    for (j = 1; j <= n; j++)
       if (j < i - m || j > i + m)
         printf("%f ", 0.0);
       else
         printf("%f ", ab[i][j-i+m+1]);
    printf("\n");
  }
}

/* ѥåƤӹɽ */
void print_symmetric_band_matrix0(int n, int m, matrix asb)
{
  int i, j;
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++)
       if (abs(j - i) > m)
         printf("%f ", 0.0);
       else {
         if (j >= i)
           printf("%f ", asb[i][j-i]);
         else /* פ i  j ؤ */
           printf("%f ", asb[j][i-j]);
       }
    printf("\n");
  }
}

/* ѥåƤӹɽ */
void print_symmetric_band_matrix1(int n, int m, matrix asb)
{
  int i, j;
  for (i = 1; i <= n; i++) {
    for (j = 1; j <= n; j++)
       if (abs(j - i) > m)
         printf("%f ", 0.0);
       else {
         if (j >= i)
           printf("%f ", asb[i][j-i+1]);
         else /* פ i  j ؤ */
           printf("%f ", asb[j][i-j+1]);
       }
    printf("\n");
  }
}

/* ͤ廰ӹɽ */
void print_upper_triangular_band_matrix1(int n, int m, matrix ub)
{
  int i, j;
  for (i = 1; i <= n; i++) {
    for (j = 1; j <= n; j++)
       if (abs(j - i) > m)
         printf("%f ", 0.0);
       else {
         if (j >= i)
           printf("%f ", ub[i][j-i+1]);
         else
           printf("%f ", 0.0);
       }
    printf("\n");
  }
}

/*  */
void mul_mm(int n, matrix ab, matrix a, matrix b)
{
  int i, j, k;
  double s;
  for (i = 1; i <= n; i++)
    for (j = 1; j <= n; j++) {
      s = 0;
      for (k = 1; k <= n; k++)
        s += a[i][k] * b[k][j];
      ab[i][j] = s;
    }
}

/* ȥ٥ȥ */
void mul_mv(int n, vector Ab, matrix A, vector b)
{
  int i, j;
  for (i = 1; i <= n; i++) {
    double s = 0;
    for (j = 1; j <= n; j++)
      s += A[i][j] * b[j];
    Ab[i] = s;
  }
}

double sqr(double x) { return x * x; }

/* Cholesky ʬ A=U^T U (A гξȾʬΤ߻) */
void cholesky(int n, matrix U, matrix a)
{
  int j, i, k;
  double s;
  for (i = 1; i <= n; i++) {
    /* U[i][i] η׻ */
    s = a[i][i];
    for (k = 1; k < i; k++)
      s -= sqr(U[k][i]);
    if (s < 0) {
      fprintf(stderr, "s < 0\n");
      exit(0);
    }
    U[i][i] = sqrt(s);
    /* U[i][j] (i < j) η׻ */
    for (j = i + 1; j <= n; j++) {
      s = a[i][j];
      for (k = 1; k < i; k++)
        s -= U[k][j] * U[k][i];
      U[i][j] = s / U[i][i];
    }
  }
}

/* Cholesky ʬ A=U^T U (A гξȾʬΤͿ) */
void cholesky_packed_v1(int n, int m, matrix U, matrix ASB)
{
  int j, i, k;
  double s;
  for (i = 1; i <= n; i++) {
    /* U[i][i] η׻ */
    s = ASB[i][1]; /* a[i][i]; */
    for (k = 1; k < i; k++)
      s -= sqr(U[k][i]);
    if (s < 0) {
      fprintf(stderr, "s < 0\n");
      exit(0);
    }
    U[i][i] = sqrt(s);
    /* U[i][j] (i < j) η׻ */
    /* ιԤ j <= n Τޤޤȡ顼ˤʤ */
    for (j = i + 1; j <= min(i + m, n); j++) {
      s = ASB[i][j-i+1]; /* a[i][j]; */
      for (k = 1; k < i; k++)
        s -= U[k][j] * U[k][i];
      U[i][j] = s / U[i][i];
    }
  }
}

/* Cholesky ʬ A=U^T U (A гξȾʬΤͿ, 0 ˤʤ) */
void cholesky_packed_v2(int n, int m, matrix U, matrix ASB)
{
  int j, i, k;
  double s;
  for (i = 1; i <= n; i++) {
    /* U[i][i] η׻ */
    s = ASB[i][1];
    for (k = max(1,i-m); k < i; k++)
      s -= sqr(U[k][i]);
    if (s < 0) {
      fprintf(stderr, "s < 0\n");
      exit(0);
    }
    U[i][i] = sqrt(s);
    /* U[i][j] (i < j) η׻ */
    for (j = i + 1; j <= min(i + m, n); j++) {
      s = ASB[i][j-i+1];
      /* ޤ i < j Ǥ롣
       * ιԤϺǽ
       *   for (k = 1; k < i; k++)
       * äU[k][j]  U[k][i] ФΤǡ
       * j-mk (k  i 꾮Τǡ j ⾮)
       * i-mk (k  i 꾮)
       * ǤʤȤʤ
       * i < j ʤΤ i-mj-m Ǥ롣
       * äơmax(1,j-m) ֤롣
       *  max(1,i-m) Ȥȡ v3 Ǥޤưʤ
       */
      for (k = max(1,j-m); k < i; k++)
        s -= U[k][j] * U[k][i];
      U[i][j] = s / U[i][i];
    }
  }
}

/*  Cholesky ʬ A=U^T U (A, U гξȾʬΤͿ) */
void cholesky_packed_v3(int n, int m, matrix UB, matrix ASB)
{
  int j, i, k;
  double s;
  for (i = 1; i <= n; i++) {
    /* U[i][i] η׻ */
    s = ASB[i][1];
    for (k = max(1,i-m); k < i; k++)
      s -= sqr(UB[k][i-k+1]);
    if (s < 0) {
      fprintf(stderr, "s < 0\n");
      exit(0);
    }
    UB[i][1] = sqrt(s);
    /* U[i][j] (i < j) η׻ */
    for (j = i + 1; j <= min(i + m, n); j++) {
      s = ASB[i][j-i+1];
      for (k = max(1,j-m); k < i; k++)
        s -= UB[k][j-k+1] * UB[k][i-k+1];
      UB[i][j-i+1] = s / UB[i][1];
    }
  }
}

/*  Cholesky ʬ A=U^T U (A, U гξȾʬΤͿ) */
void cholesky_packed_v4(int n, int m, matrix ASB)
{
  int j, i, k;
  double s;
  for (i = 1; i <= n; i++) {
    /* U[i][i] η׻ */
    s = ASB[i][1];
    for (k = max(1,i-m); k < i; k++)
      s -= sqr(ASB[k][i-k+1]);
    if (s < 0) {
      fprintf(stderr, "s < 0\n");
      exit(0);
    }
    ASB[i][1] = sqrt(s);
    /* U[i][j] (i < j) η׻ */
    for (j = i + 1; j <= min(i + m, n); j++) {
      s = ASB[i][j-i+1];
      for (k = max(1,j-m); k < i; k++)
        s -= ASB[k][j-k+1] * ASB[k][i-k+1];
      ASB[i][j-i+1] = s / ASB[i][1];
    }
  }
}

/* ѹ򷸿ȤϢΩ켡 */
/* Ux=b  */
void solve_uxb(int n, matrix u, vector b)
{
  int i, j;
  for (i = n; i >= 1; i--) {
    for (j = i + 1; j <= n; j++)
      b[i] -= u[i][j] * b[j];
    b[i] /= u[i][i];
  }
}

/* Ux=b  */
void solve_band_uxb(int n, int m, matrix UB, vector b)
{
  int i, j;
  for (i = n; i >= 1; i--) {
    for (j = i + 1; j <= min(i+m,n); j++)
      b[i] -= UB[i][j-i+1] * b[j];
    b[i] /= UB[i][1];
  }
}

/* U^T x=b  */
void solve_band_utxb(int n, int m, matrix UB, vector b)
{
  int i, j;
  for (i = 1; i <= n; i++) {
    for (j = max(i-m,1); j < i; j++)
      b[i] -= UB[j][i-j+1] * b[j];
    b[i] /= UB[i][1];
  }
}

int main()
{
  int i,j, n,m;
  matrix U, Ut, A, ASB, u, UB;
  vector x, b;

  n = 8; m = 3;
  /* Ѥΰ */
  U = new_matrix(n + 1, n + 1);
  u = new_matrix(n + 1, n + 1);
  Ut = new_matrix(n + 1, n + 1);
  A = new_matrix(n + 1, n + 1);
  ASB = new_matrix(n + 1, (m + 1) + 1);
  UB  = new_matrix(n + 1, (m + 1) + 1);
  x = new_vector(n + 1);
  b = new_vector(n + 1);
  /* Ѱ (Ѥ) */
  for (i = 1; i <= n; i++) {
    for (j = 1; j <= n; j++)
      Ut[j][i] = U[i][j] = 0;
    for (j = i; j <= min(i + m, n); j++)
      Ut[j][i] = U[i][j] = drandom();
  }
  mul_mm(n, A, Ut, U);
  /* A Ȱ U ɽ */
  printf("U=\n"); print_matrix1(n, n, U);
  printf("a=\n"); print_matrix1(n, n, A);

  /* A  ASB ˵ͤ */
  pack_symmetric1(n, m, ASB, A);
  printf("ѥå A=\n"); print_matrix1(n, m+1, ASB);
  printf("̾ɽ\n");
  print_symmetric_band_matrix1(n, m, ASB);

  /* ޤѤ Cholesky ʬ */
  cholesky(n, u, A);
  printf("A ʬ򤷤Ƶޤä u=\n"); print_matrix1(n, n, u);

  /* A ͤϤƤߤ */
  cholesky_packed_v1(n, m, U, ASB);
  printf("cholesky_packed_v1() ǵޤä U=\n");
  print_matrix1(n, n, U);

  /* A ͤϤƤߤ, ̵̤ʷ׻Ϥʤ */
  cholesky_packed_v2(n, m, U, ASB);
  printf("cholesky_packed_v2() ǵޤä U=\n");
  print_matrix1(n, n, U);

  /* A ͤϤơ̤֤ͤ */
  cholesky_packed_v3(n, m, UB, ASB);
  printf("cholesky_packed_v3() ǵޤä U=\n");
  print_upper_triangular_band_matrix1(n, m, UB);

  /* A  U 񤭤֤ */
  for (i = 1; i <= n; i++)
    for (j = 1; j <= m + 1; j++)
      UB[i][j] = ASB[i][j];
  cholesky_packed_v4(n, m, UB);
  printf("cholesky_packed_v4() ǵޤä U=\n");
  print_upper_triangular_band_matrix1(n, m, UB);

  /* ϢΩ1 */
  for (i = 1; i <= n; i++)
    x[i] = i;
  mul_mv(n, b, A, x);
  solve_band_utxb(n, m, UB, b);
  solve_band_uxb(n, m, UB, b);

  printf("x=\n");
  print_vector1(n, b);
  return 0;
}
